{ ljr, 3/20/89, vb3 }

{$S-,R-,V-,I-,B-,F-}
{$I TPDEFINE.INC}      { MAJOR MODIFACATION! to do bios writes only }

program VBrowse;
  {-Virtual file browser}
uses
  Dos,                       {standard DOS/BIOS routines}
  TpString,                  {Turbo Professional string handling routines}
  TpDos,                     {Turbo Professional disk/file management, etc.}
  TpCrt,                     {Turbo Professional CRT unit}
  TpVArray;                  {Turbo Professional virtual arrays}

const
  {--------------- customization section ------------}
  MaxLines = 64000;          {maximum number of lines}
  MaxLineSize = 128;         {maximum size of a line}
  HScrollDelta = 20;         {^Left/^Right scroll horizontally this many columns}
  SwapName : string[64] =    {file used by TPVARRAY to store array elements}
  'VB-SWAP';
  ExpandTabs = True;         {if True, tabs are expanded to spaces}
  TabSize = 4;               {distance between tab stops when expanding tabs}
  IgnoreCase : Boolean = True; {ignore case when searching?}

  {messages}
  FileOpenError = 'Error while trying to open file';
  FileReadError = 'Error while trying to read file';
  FileCloseError = 'Error while trying to close file';
  SwapWriteError = 'Error while writing to swap file';
  SwapReadError = 'Error while reading from swap file';
  FileNotFound = 'File not found';
  FilePrompt = 'File: ';
  FindPrompt = 'Find: ';
  GotoPrompt = 'Goto line: ';
  IgnoreCasePrompt = 'Ignore case?';
  StringNotFound = 'String not found';
  InsufficientMemory = 'Insufficient memory';
  SearchingMsg = 'Searching...';
  PressAnyKeyPrompt = '. Press any key...';
  AbortPrompt = '. Press any key to return to DOS.';

  {screen coordinates}
  InsideRow = 1;             {first row in viewing windew}
  InsideCol = 1;             {first column in viewing windew}
  PromptCol = 1;             {column for prompt messages}
  FNameCol = 2;              {column for filename}
  LinesCol = 17;             {column for current line display}
  ColCol = 39;               {column for current column display}
  TotalCol = 49;             {column for total lines display}
  BytesCol = 63;             {column for total bytes display}

  {help}
  HelpLines = 9;            {number of lines in help screen}
  HelpLength = 79;           {length of strings in help screen}
  HelpText : array[1..HelpLines] of string[HelpLength] = (
'1 HELP for VBROWSE ver 1.00 (bios calls)  press <UP>, <DOWN>, <ESC> to exit ',
'2 Scroll up      <Up>       <CtrlE>   Scroll down     <Down>      <CtrlX>  ',
'3 Scroll left    <Left>     <CtrlS>   Scroll right    <Right>     <CtrlD>  ',
'4 Scroll left 20 <CtrlLeft> <CtrlA>   Scroll right 20 <CtrlRight> <CtrlF>  ',
'5 Page up        <PgUp>     <CtrlR>   Page down       <PgDn>      <CtrlC>  ',
'6 Top of file    <^PgUp>    <Home>    Bottom of file  <^PgDn>     <End>    ',
'7 Find string    <CtrlQ>F             Repeat find     <CtrlL>              ',
'8 Set marker 0-9 <CtrlK>n             Goto marker 0-9 <CtrlQ>n             ',
'9 Goto line      <CtrlQ>L             Exit            <AltX>      <Esc>    ');
type
  Line = string[MaxLineSize]; {a line of text}
var
  LineArray : TpArray;       {pointer to the virtual array}
  FName : string[64];        {file being browsed}
  FNameNP : string[13];      {" -- with no path}
  TotalLineStr : string[7];  {string representing line count}
  TotalByteStr : string[10]; {string representing byte count}
  CurrentCol : Integer;      {column being displayed at left edge of screen}
  CurrentLine,               {line number being displayed at top of windew}
  LastLine,                  {last line that can be displayed}
  NumRows,                   {number of rows in viewing windew}
  TotalLines,                {total # of lines in the file}
  LineWidth,                 {width of the browsing windew}
  SaveMin,                   {save windmin}
  SaveMax,                   {save windmax}
  StatusLine : Word;         {row for status line}
  OrigColor,                 {Original color }
  Reverse,                   {reverse video attribute}
  Bright,                    {bright video attribute}
  Highlight,                 {highlight video attribute}
  Dim : Byte;                {dim video attribute}
  TotalBytes : LongInt;      {total # of bytes in the file}
  SearchString : Line;       {last string searched for}
  LastCurrentLine : Word;    {used for search operations}
  SuppressStatus : Boolean;  {if False, status line not drawn}
  Markers : array[0..9] of
    record
      IsSet : Boolean;       {Is marker set?}
      Row, Col : Word;       {line and column}
    end;
  ch : char;
  Forever : boolean;

{------------------------------------------------------------}
  procedure CleanUp;
    {-Delete swap file and restore screen}
  begin
    {get rid of the swap file}
    DisposeA(LineArray, True);
    if ErrorA <> 0 then
      {do nothing} ;
    {restore the screen}
    TextAttr := OrigColor;
    ClrScr;
    NormalCursor;
  end;

  procedure SetAttributes;
    {-Set video attribute variables based on the current video mode}
  var
    MonoColors : Boolean;
  begin
    {we're doing our own color mapping}
    MapColors := False;

    {set video attributes}
    case CurrentMode of
      2, 7 :
        MonoColors := True;
      else
        MonoColors := False;
    end;

    if MonoColors then begin
      Bright := $F;          {white on black}
      Dim := $7;             {light gray on black}
      Highlight := $70;      {black on light gray}
      Reverse := $70;        {black on light gray}
    end
    else begin
      Bright := $1e;         {yellow on blue}
      Dim := $1e;            {yellow on blue}
      Highlight := $74;      {red on lightgray}
      Reverse := $6e;        {yellow on brown}
    end;
    TextAttr := Dim;
  end;

  function PaddedWordStr(W, Width : Word) : string;
    {-Return W as a string, padded to the specified width}
  var
    S : string;
  begin
    FillChar(S[1], Width, ' ');
    Str(W, S);
    S[0] := Chr(Width);
    PaddedWordStr := S;
  end;

  procedure FWrite (Msg : string; Row, Col : Byte);
  begin
    GotoXY(Col,Row);
    TextAttr := Dim;
    Write(Msg);
  end;

  procedure FWriteS (Msg : string; Col : Byte); { for STATUS/HELP line }
  begin
    Window(1,25,80,1);
    GotoXY(Col,25);
    TextAttr := Reverse;
    Write(Msg);
  end;

  procedure ClearStatusLine;
    {-Clear the status line}
  begin
    FWriteS (Pad('', 79), InsideCol);
  end;

  procedure DrawStatusLine;
    {-Display status information}
  var
    S : string;
    N : Word;
  begin
    {exit if status line suppressed}
    if SuppressStatus then
      Exit;

    {show filename}
    FWriteS (FNameNP, FNameCol);

    {show lines on screen}
    N := CurrentLine+Pred( NumRows);
    if N > TotalLines then
      N := TotalLines;
    S := 'Lines: '+Long2Str(CurrentLine)+' - '+Long2Str(N);

    FWriteS (Pad(S, 20), LinesCol);

    {show leftmost column}
    FWriteS ('Col: '+PaddedWordStr(CurrentCol, 3), ColCol);

    {show total lines}
    FWriteS ('Total: '+TotalLineStr, TotalCol);

    {show total bytes}
    FWriteS ('Bytes: '+TotalByteStr, BytesCol);
  end;

  procedure ReadMyStr(var S : string; R, C, W : byte; var Escaped : boolean);
  var
    SLen : byte absolute S;
    ch : char;
    CurrentCol,
    LastCol,
    I : byte;
  begin
    Escaped := false;
    I       := SLen;
    CurrentCol := C;
    LastCol := C + W - 1;
    repeat
      ch := ReadKey;
      case ch of
        #27 : begin
                SLen := 0;
                Escaped := true;
                Exit;
              end;
        #13 : Exit;
      end; { case }
      if ch in [' '..'~'] then
        begin
          if CurrentCol < LastCol then
            begin
              inc (I);
              S[I] := ch;
              inc(SLen);
              Write(ch);
              inc(CurrentCol);
            end
          else
        end
      else
        if ch = #08 then
          begin
            if CurrentCol > C then
              begin
                dec(I);
                dec(SLen);
                dec(CurrentCol);
                GotoXY(CurrentCol,R);
                Write(' ');
                GotoXY(CurrentCol,R);
              end;
          end
        else Write(#7);
    until forever;
  end;

  procedure Prompt(Msg : string; var Escaped : Boolean; var S : string);
    {-Prompt for a string}
  var
    Row, Col, Width : byte;
  begin
    ClearStatusLine;
    FWriteS (Msg, PromptCol);
    Row := WhereY;
    Col := WhereX;
    Width := 50; { arbitrary set limit }
    NormalCursor;
    ReadMyStr(S, Row, Col, Width, Escaped);
    HiddenCursor;
    ClearStatusLine;
    DrawStatusLine;
  end;

  procedure PromptWord(Msg : string; Lo, Hi : Word; var Escaped : Boolean; var W : Word);
    {-Prompt for a word}
  var
    Row, Col, Width : byte;
    LongNum : word;
    NumStr : string;
    Result : integer;
    okay : boolean;
  begin
    okay := false;
    ClearStatusLine;
    FWriteS (Msg, PromptCol);
    Row := WhereY;
    Col := WhereX;
    Width := 5;     { arbitrary set limit to number of digits }
    NormalCursor;
    repeat
      NumStr[0] := #0;
      ReadMyStr(NumStr ,Row ,Col ,Width ,Escaped );
      val(NumStr,LongNum,Result);
      if LongNum < Hi then
       begin
         okay := true;
         CurrentLine := LongNum;
       end;
    until okay;
     HiddenCursor;
     ClearStatusLine;
     DrawStatusLine;
  end;

  function PromptYesOrNo(Msg : string; Default : Char) : Boolean;
    {-Get a response to a yes-no question}
  begin
    ClearStatusLine;
    FWriteS (Msg, PromptCol);
    NormalCursor;
    repeat ch := ReadKey; until ch in [ 'Y','y','N','n' ];
    Default := ch;
    if ch in [ 'Y','y' ] then PromptYesOrNo := true
                         else PromptYesOrNo := false;
    HiddenCursor;
    ClearStatusLine;
    DrawStatusLine;
  end;

  procedure PromptMessage(Msg : string);
    {-Display a message on the prompt line}
  begin
    ClearStatusLine;
    FWriteS (Msg, PromptCol);
  end;

  procedure Abort(Msg : string);
    {-Show an error message and halt}
  var
    I : Word;
  begin
    PromptMessage(Msg+AbortPrompt);
    I := ReadKeyWord;
    CleanUp;
    Halt(1);
  end;

  procedure PressAnyKey(Msg : string);
    {-Display a message on the prompt line and wait for keypress}
  var
    I : Word;
  begin
    PromptMessage(Msg+PressAnyKeyPrompt);
    I := ReadKeyWord;
    ClearStatusLine;
    DrawStatusLine;
  end;

  procedure GetLine(Line : Word; var S : Line);
    {-Get the specified line into S}
  begin
    RetA(LineArray, Pred(Line), 0, S);
    if ErrorA <> 0 then
      Abort(SwapReadError);
  end;

  procedure DrawTheLine(S : String; Row,Col : Byte);
  begin
    SaveMin := WindMin;
    SaveMax := WindMax;
    Window(1,1,80,25);
    FWrite (S, Row + Hi(WindMin), Col + Lo(WindMin));
    Window (Succ (Lo (SaveMin)),
            Succ (Hi (SaveMin)),
            Succ (Lo (SaveMax)),
            Succ (Hi (SaveMax)));
  End;

  procedure DrawLine(Row : Byte);
    {-Draw the line that goes at the specified Row (window-relative)}
  var
    S : Line;
    SLen : Byte absolute S;
    LineNum : Word;
  begin
    {calculate line number and get the string}
    LineNum := CurrentLine+Pred(Row);
    if LineNum <= TotalLines then
      GetLine(LineNum, S)
    else
      SLen := 0;

    {adjust for current column}
    if SLen < CurrentCol then
      SLen := 0
    else
      if CurrentCol <> 1 then begin
        Dec(SLen, Pred(CurrentCol));
        Move(S[CurrentCol], S[1], SLen);
      end;

    {pad the end of the line with blanks if necessary}
    if LineWidth > SLen then
      FillChar(S[Succ(SLen)], LineWidth-SLen, ' ');
    SLen := LineWidth;

    {draw the line}
    DrawTheLine(S, Row, 1);
  end;

  procedure DrawFullScreen;
    {-Redraw the screen}
  var
    Row : Byte;
  begin
    {make sure the column is a legal value}
    if CurrentCol < 1 then
      CurrentCol := 1;

    {draw the individual lines}
    for Row := 1 to 24 do
      DrawLine(Row);
  end;

  procedure ReadNewFile(FN : string);
    {-Read the contents of the file into the array}
  const
    BufSize = 8192;
  var
    F : Text;
    Buffer : array[1..BufSize] of Byte;
    S : string;
  begin                      {ReadFile}
    {clean up filenames}
    FName := StUpcase(CleanPathName(FN));
    FNameNP := Pad(JustFilename(FName), 13);

    {reset column/line counters}
    CurrentCol := 1;
    CurrentLine := 1;
    TotalLines := 0;

    {assign the filename and allocate a buffer}
    Assign(F, FName);
    SetTextBuf(F, Buffer, BufSize);

    {open the file}
    Reset(F);
    if IoResult <> 0 then
      Abort(FileOpenError);

    {get size of file}
    TotalBytes := TextFileSize(F);
    if TotalBytes = -1 then
      Abort(FileReadError);
    TotalByteStr := Long2Str(TotalBytes);

    {show what we're doing}
    PromptMessage('Reading '+FName+'...');

    {read the file into the array}
    while not Eof(F) and (TotalLines < MaxLines) do begin
      {read next line}
      ReadLn(F, S);
      if IoResult <> 0 then
        Abort(FileReadError);

      {expand tabs if desired}
      if ExpandTabs then
        S := DeTab(S, TabSize);

      {make sure line isn't too long}
      if Length(S) > MaxLineSize then
        S[0] := Chr(MaxLineSize);

      {plug the line into the array}
      SetA(LineArray, TotalLines, 0, S);
      if ErrorA <> 0 then
        Abort(SwapWriteError);

      Inc(TotalLines);

      {update the screen when we have a screenful}
      if TotalLines = NumRows then
        DrawFullScreen;
    end;

    {convert total lines to a string}
    TotalLineStr := PaddedWordStr(TotalLines, 5);

    {set LastLine}
    if TotalLines < NumRows then begin
      LastLine := TotalLines;
      {screen not drawn yet--do it now}
      DrawFullScreen;
    end
    else
      LastLine := Succ(TotalLines-NumRows);

    {get rid of Reading... message}
    ClearStatusLine;

    {close the file}
    Close(F);
    if IoResult <> 0 then
      Abort(FileCloseError);

    {clear markers}
    FillChar(Markers, SizeOf(Markers), 0);

    {enable the status line}
    SuppressStatus := False;
  end;

  procedure ScrollUp;
    {-Scroll up one line}
  begin
    if CurrentLine > 1 then begin
      Dec(CurrentLine);
      TextAttr := Dim;
      ScrollWindowDown(1,1,80,24,1);
      DrawLine(1);
    end;
  end;

  procedure ScrollDown;
    {-Scroll down one line}
  begin
    if CurrentLine < TotalLines then begin
      Inc(CurrentLine);
      TextAttr := Dim;
      ScrollWindowUp(1,1,80,24,1);
      DrawLine(24);
    end;
  end;

  procedure ScrollLeft;
    {-Scroll to the left 20 columns}
  begin
    if CurrentCol > 1 then begin
      if CurrentCol >= HScrollDelta then
        Dec(CurrentCol, HScrollDelta)
      else
        CurrentCol := 1;
      DrawFullScreen;
    end;
  end;

  procedure ScrollRight;
    {-Scroll to the right 20 columns}
  begin
    if CurrentCol < MaxLineSize then begin
      Inc(CurrentCol, HScrollDelta);
      if CurrentCol > MaxLineSize then
        CurrentCol := MaxLineSize;
      DrawFullScreen;
    end;
  end;

  procedure Left;
    {-Scroll to the left 1 column}
  begin
    if CurrentCol > 1 then begin
      Dec(CurrentCol);
      DrawFullScreen;
    end;
  end;

  procedure Right;
    {-Scroll to the right 1 column}
  begin
    if CurrentCol < MaxLineSize then begin
      Inc(CurrentCol);
      DrawFullScreen;
    end;
  end;

  procedure PageUp;
    {-Scroll up one full screen}
  var
    I : Integer;
  begin
    if CurrentLine > 1 then begin
      I := CurrentLine-NumRows;
      if I > 1 then
        CurrentLine := I
      else
        CurrentLine := 1;
      DrawFullScreen;
    end;
  end;

  procedure PageDown;
    {-Scroll down one full screen}
  var
    I : Word;
  begin
    if CurrentLine < LastLine then begin
      I := CurrentLine+NumRows;
      if I < TotalLines then
        CurrentLine := I
      else
        CurrentLine := TotalLines;
      DrawFullScreen;
    end;
  end;

  procedure TopOfFile;
    {-Go to first line in file}
  begin
    if (CurrentLine > 1) or (CurrentCol > 1) then begin
      CurrentLine := 1;
      CurrentCol := 1;
      DrawFullScreen;
    end;
  end;

  procedure EndOfFile;
    {-Go to last line in file}
  begin
    if (CurrentLine < LastLine) or (CurrentCol > 1) then begin
      CurrentLine := LastLine;
      CurrentCol := 1;
      DrawFullScreen;
    end;
  end;

  procedure GotoLine;
    {-Process goto line command}
  var
    Escaped : Boolean;
  begin
    PromptWord(GotoPrompt, 1, TotalLines, Escaped, CurrentLine);
    if not Escaped then
      {redraw the screen}
      DrawFullScreen;
  end;

  procedure FindPrimitive;
    {-Primitive routine to find a string}
  const
    LCount : Word = 0;
  var
    BT : BTable;
    Index, I, J : Word;
    SearchFor, L : Line;
    Found : Boolean;
  begin
    {show status message}
    PromptMessage(SearchingMsg);

    {set up for search}
    if IgnoreCase then
      SearchFor := StUpcase(SearchString)
    else
      SearchFor := SearchString;
    BMMakeTable(SearchFor, BT);

    {start searching at current line, or line after last find}
    if LastCurrentLine <> CurrentLine then
      LCount := Pred(CurrentLine);

    Found := False;
    while (LCount < TotalLines) and not Found do begin
      {get next line}
      Inc(LCount);
      GetLine(LCount, L);

      {skip lines that are too short}
      if Length(L) >= Length(SearchFor) then begin
        {search for a match}
        if IgnoreCase then
          Index := BMSearchUC(L[1], Length(L), BT, SearchFor)
        else
          Index := BMSearch(L[1], Length(L), BT, SearchFor);

        {search routines return $FFFF if string not found}
        if (Index <> $FFFF) then begin
          Found := True;
          Inc(Index);
        end;
      end;
    end;

    {if not found, show error message}
    if not Found then begin
      PressAnyKey(StringNotFound);
      Exit;
    end;

    {set current line so that the line is clearly visible}
    if LCount >= (CurrentLine+NumRows) then
      CurrentLine := LCount-2;
    LastCurrentLine := CurrentLine;

    {set current column so that the line is clearly visible}
    I := Index+Length(SearchString)+2;
    if (I-Pred(CurrentCol)) > LineWidth then
      CurrentCol := I-LineWidth;
    if Integer(Index)-2 < CurrentCol then
      CurrentCol := Index-2;

    {redraw the screen}
    DrawFullScreen;

    {redraw the status line}
    ClearStatusLine;
    SuppressStatus := False;
    DrawStatusLine;

    {highlight the search string}
    I := InsideRow+(LCount-CurrentLine);
    J := InsideCol+(Index-CurrentCol);

    ChangeAttribute(Length(SearchString), I, J, Highlight);

    while not KeyPressed do {};

    ChangeAttribute(Length(SearchString), I, J, Dim);
  end;

  procedure FindString;
    {-Prompt for a string and find it}
  const
    YorN : array[Boolean] of Char = ('N', 'Y');
  var
    Escaped : Boolean;
  begin
    {keep the status line from flashing}
    SuppressStatus := True;

    {get the search string}
    SearchString[0] := #0;
    Prompt(FindPrompt, Escaped, SearchString);

    {exit if Escape pressed or no string}
    if Escaped or (Length(SearchString) = 0) then
      Exit;

    {case sensitive search?}
    IgnoreCase := PromptYesOrNo(IgnoreCasePrompt, YorN[IgnoreCase]);

    {new search string, start from current line}
    LastCurrentLine := 0;

    {do the search operation}
    FindPrimitive;
  end;

  procedure FindLast;
    {-Find the last search string again, if any}
  begin
    {if there's a search string, use it again}
    if Length(SearchString) <> 0 then
      FindPrimitive;
  end;

  procedure SetMarker(N : Byte);
    {-Set a marker}
  begin
    with Markers[N] do begin
      IsSet := True;
      Col := CurrentCol;
      Row := CurrentLine;
    end;
  end;

  procedure JumpMarker(N : Byte);
    {-Jump to a marker}
  begin
    with Markers[N] do
      if IsSet then begin
        {reset current line and column}
        CurrentCol := Col;
        CurrentLine := Row;

        {redraw}
        DrawFullScreen;
      end
      else
        PressAnyKey('Marker '+Chr(N+Ord('0'))+' not set');
  end;

  procedure Help;
    {-Show help screen}
  var
    I,key : Word;
    ch    : char absolute key;
  begin
    I := 1;
    ClearStatusLine;
    repeat
      begin
        FWriteS (HelpText[I], 1);
        key := ReadKeyWord;
        case (hi(key)) of
          $47, $48, $49 : case I of
                            1            : {};
                            2..HelpLines : dec(I);
                          end;
          $4F, $50, $51 : if (I < HelpLines) then inc(I);
        end;
      end;
    until (ch = #27);
    ClearStatusLine;
  end;

  procedure Browse;
    {-Browse through the file}
  const
    Esc = #27;
  var
    ChWord : Word;
    Ch : Char absolute ChWord;
    Done : Boolean;

    function WordStarCommand(Ch : Char) : Char;
      {-Return ^C, 'C', or 'c' as ^C, etc.}
    begin
      Ch := Upcase(Ch);
      case Ch of
        'A'..'Z' : WordStarCommand := Chr(Ord(Ch)-64);
      else WordStarCommand := Ch;
      end;
    end;

  begin
    Done := False;
    repeat
      {redraw the status line}
      SuppressStatus := False;
      DrawStatusLine;

      {get next key}
      ChWord := ReadKeyWord;

      {check for extended scan codes}
      if Ch = #0 then case Hi(ChWord) of
        $00,                 {CtrlBreak}
        $2D : Done := True;  {AltX}
        $3B : Help;          {F1 -> Help}
        $47,                 {Home}
        $84 : TopOfFile;     {^PgUp}
        $48 : ScrollUp;      {Up}
        $49 : PageUp;        {PgUp}
        $4B : Left;          {Left}
        $4D : Right;         {Right}
        $4F,                 {End}
        $76 : EndOfFile;     {^PgDn}
        $50 : ScrollDown;    {Down}
        $51 : PageDown;      {PgDn}
        $73 : ScrollLeft;    {^Left - scroll left 20}
        $74 : ScrollRight;   {^Right - scroll right 20}
      end
      else case Ch of
        {regular keys}
        ^K :                 {WordStar ^K commands}
          begin
            {get next key}
            ChWord := TpCrt.ReadKeyWord;
            case WordStarCommand(Ch) of
              '0'..'9' :     {set marker}
                SetMarker(Ord(Ch)-Ord('0'));
            end;
          end;
        ^Q :                 {WordStar ^Q commands}
          begin
            {get next key}
            ChWord := TpCrt.ReadKeyWord;
            case WordStarCommand(Ch) of
              '0'..'9' :     {jump to marker}
                JumpMarker(Ord(Ch)-Ord('0'));
              ^R : TopOfFile; {top of file}
              ^C : EndOfFile; {end of file}
              ^F : FindString; {search command}
              ^L : GotoLine; {goto line command}
            end;
          end;
        ^E,
        ^W : ScrollUp;       {scroll up}
        ^Z,
        ^X : ScrollDown;     {scroll down}
        ^S : Left;           {scroll left 1}
        ^D : Right;          {scroll right 1}
        ^A : ScrollLeft;     {scroll left 20}
        ^F : ScrollRight;    {scroll right 20}
        ^R : PageUp;         {page up}
        ^C : PageDown;       {page down}
        ^L : FindLast;       {find last search string}
        Esc : Done := True;  {Escape}
      end;
    until Done;
  end;

  procedure FetchFile;
  var
    OK,
    Escaped : boolean;
  begin
    {get filename}
    FName := ParamStr(1);
    OK := (length(FName) <> 0);
    repeat
      if not OK then
        begin
          Prompt(FilePrompt, Escaped, FName);
          if Escaped then
            begin
              CleanUp;
              Halt(0);
            end;
        end;

      { look for the file }
      OK := ExistOnPath(FName, FName);
      if not OK then
        PressAnyKey(FileNotFound);
    until OK;

    {add path identifier to swap file name, if any}
    if ParamCount >= 2 then
      SwapName := JustPathName(ParamStr(2))+SwapName;

    {read in the file}
    ReadNewFile(FName);
  end;

  procedure Create_Array;
  begin
    {create the array}
    WriteError := False;
    HaltOnError := False;
    MakeA(LineArray, MaxLines, 1, SizeOf(Line), SwapName, MaxAvail shr 1);
    if ErrorA <> 0 then
      {presumably not enough memory}
      Abort(InsufficientMemory);
  end;

  procedure Initialize;
    {-Initialization code}
  var
    Escaped, OK : Boolean;
    Regs : Registers;
  begin
    Regs.ah := 8;
    Regs.bh := 0;
    Intr ($10,Regs);
    OrigColor := Regs.ah;      { get user's original color setup }

    ClrScr;
    {initialize variables}
    DirectVideo := false;
    BiosScroll := true;       { allow as background task }
    FileMode := $42;           {allow to run on local area network }
    CheckBreak := False;
    Forever := false;
    SuppressStatus := True;
    SearchString := '';
    FNameNP := '';
    NumRows := ScreenHeight-1;
    LineWidth := ScreenWidth-1;
    StatusLine := 1;

    {set up video attributes}
    SetAttributes;

    {make sure we can run under DesqView/TaskView}
    DetectMultitasking := True;
    ReinitCrt;

    {see if help is requested}
    if Pos('?', ParamStr(1)) <> 0 then begin
      WriteLn('File Browser (using bios writes...ie no direct screen writes).'^M^J);
      WriteLn('Usage: VBROWSE FileName [Option]'^M^J);
      WriteLn('where the option is "d:\path" (the Drive and directory for swap file).');
      Halt(0);
    end;
    ClearStatusLine;
    {hide the cursor}
    HiddenCursor;
  end;

{***************** M A I N *********************}

begin
  {get command line parameters, if any, and initialize}
  Initialize;

  {make a virtual array}
  Create_Array;

  {get the file to browse}
  FetchFile;

  {browse through the file}
  Browse;

  {clean up the screen}
  CleanUp;
end.
